#include <stdlib.h>
#include <stdio.h>

#define FILE_XX "z80op_xx.c"
#define FILE_ED "z80op_ed.c"
#define FILE_DD "z80op_dd.c"
#define FILE_FD "z80op_fd.c"
#define FILE_CB "z80op_cb.c"
#define FILE_DDCB "z80op_ddcb.c"
#define FILE_FDCB "z80op_fdcb.c"

#define SUFFIX_XX "xx"
#define SUFFIX_ED "ed"
#define SUFFIX_DD "dd"
#define SUFFIX_FD "fd"
#define SUFFIX_CB "cb"
#define SUFFIX_DDCB "ddcb"
#define SUFFIX_FDCB "fdcb"

#define d3bit(i) ((i)&0x07)          /* .....210 */
#define d4bit(i) ((i)&0x0f)          /* ....3210 */
#define m3bit(i) (((i)>>3) & 0x07)   /* ..210... */
#define m2bit(i) (((i)>>4) & 0x03)   /* ..10.... */
#define u2bit(i) (((i)>>6) & 0x03)   /* 10...... */

static FILE *open_srcfile(char *filename, char *suffix)
{
	FILE *fp;

	if ((fp = fopen(filename, "w")) == NULL) {
		fprintf(stderr, "Can't create file: %s\n", filename);
		exit(1);
	}
	fprintf(fp, "#include <stdio.h>\n");
	fprintf(fp, "#include \"z80.h\"\n");
	fprintf(fp, "#include \"z80operand.h\"\n\n");
	return fp;
}

static int close_srcfile(FILE *fp, char *suffix)
{
	int i;

	fprintf(fp, "\nint (*(instr_%s[]))() = {\n", suffix);

	for (i = 0; i < 256; i++) {
		fprintf(fp, "\texec_%s%02x,\n", suffix, i);
	}
	fprintf(fp, "};\n");

	return fclose(fp);
}

static int print_func_header(FILE *fp, char *suffix, int code)
{
	fprintf(fp, "int exec_%s%02x()\n", suffix, code);
	fprintf(fp, "{\n");
	return 1;
}

static int print_func_footer(FILE *fp, int clockadd)
{
	fprintf(fp, "\treturn %d;\n", clockadd);
	fprintf(fp, "}\n");
	return 1;
}

static int print_func_footer_con(FILE *fp, char *condition, 
				 int clockadd1, int clockadd2)
{
	fprintf(fp, "\treturn (%s)? %d : %d;\n", condition, clockadd1, clockadd2);
	fprintf(fp, "}\n");
	return 1;
}

static int print_func_footer_str(FILE *fp, char *string)
{
	fprintf(fp, "\treturn %s;\n", string);
	fprintf(fp, "}\n");
	return 1;
}

static char *reg8s[] = {"B", "C", "D", "E", "H", "L", NULL, "A", "R", "I"};
static char *reg8s_ix[] = {"B", "C", "D", "E", "IXH", "IXL", NULL, "A", "I", "R"};
static char *reg8s_iy[] = {"B", "C", "D", "E", "IYH", "IYL", NULL, "A", "I", "R"};
static char *reg16s[] = {"BC", "DE", "HL", "SP", "AF", "AF2"};
static char *reg16s_ix[] = {"BC", "DE", "IX", "SP", "AF", "AF2"};
static char *reg16s_iy[] = {"BC", "DE", "IY", "SP", "AF", "AF2"};
static char *flags[] = {"NZ", "Z", "NC", "C", "PO", "PE", "P", "M"};
static char *flagconds[] = {
	"!bitFlg(Z)", "bitFlg(Z)",
	"!bitFlg(C)", "bitFlg(C)",
	"!bitFlg(PV)", "bitFlg(PV)",
	"!bitFlg(S)", "bitFlg(S)",
};

#define NONE_T    0
#define REG8_T    1   /* R8 */
#define REG16_T   2   /* R16 */
#define REGREF_T  3   /* RM */
#define IDXREF_T  4   /* IDX */
#define CONST_T   5   /* N */
#define VALUE8_T  6   /* N */
#define VALUE16_T 7   /* N */
#define VALREF_T  8   /* M */
#define VAL8REF_T 9   /* M */
#define REGRI_T   10  /* RI */
#define FLAG_T    11  /* FLG */

static char *midx[]={"", "_R8", "_R16", "_RM", "_IDX", "_N", "_N", "_N", "_M", "_M", "_RI", "_FLG"};
static int needfetch[]={0, 0, 0, 0, 1, 0, 1, 2, 2, 1, 0, 0};

static int print_arg(FILE *fp, int argt, int argv, char **reg8s_,
		     char **reg16s_, char *vartmp)
{
	switch(argt) {
	case REG8_T: case REGRI_T:
		fprintf(fp, "%s", reg8s_[argv]);
		break;
	case REG16_T: case REGREF_T:
		fprintf(fp, "%s", reg16s_[argv]);
		break;
	case IDXREF_T:
		fprintf(fp, "%s,%s", reg16s_[argv], vartmp);
		break;
	case CONST_T:
		fprintf(fp, "%d", argv);
		break;
	case VALUE8_T: case VALUE16_T: case VALREF_T: case VAL8REF_T:
		fprintf(fp, "%s", vartmp);
		break;
	case FLAG_T:
		fprintf(fp, "%s", flags[argv]);
		break;
	}
	return 1;
}

static int print_arg_for_code(FILE *fp, int argt, int argv, char **reg8s_,
			      char **reg16s_)
{
	switch(argt) {
	case REG8_T:
	case REGRI_T:
		fprintf(fp, "%s", reg8s_[argv]);
		break;
	case REG16_T:
		fprintf(fp, "%s", (argv != 5)? reg16s_[argv] : "AF'");
		break;
	case REGREF_T:
		fprintf(fp, "(%s)", reg16s_[argv]);
		break;
	case IDXREF_T:
		fprintf(fp, "(%s+%%02xH)", reg16s_[argv]);
		break;
	case CONST_T:
		if (argv >= 8) fprintf(fp, "%02xH", argv);
		else fprintf(fp, "%d", argv);
		break;
	case VALUE8_T:
		fprintf(fp, "%%02xH");
		break;
	case VALUE16_T:
		fprintf(fp, "%%04xH");
		break;
	case VALREF_T:
		fprintf(fp, "(%%04xH)");
		break;
	case VAL8REF_T:
		fprintf(fp, "(%%02xH)");
		break;
	case FLAG_T:
		fprintf(fp, "%s", flags[argv]);
		break;
	}
	return 1;
}

static int print_func_body_with_ld(FILE *fp, char *op, int arg1t, int arg1v,
				   int arg2t, int arg2v, int ddfd, int ldarg) {
	char **reg8s_ = (ddfd == 0)? reg8s : ((ddfd == 1)? reg8s_ix : reg8s_iy);
	char **reg16s_ = (ddfd == 0)? reg16s : ((ddfd == 1)? reg16s_ix : reg16s_iy);
	int isddfd = (ddfd > 0);
	int withld = ((ldarg >= 0)&&(ldarg != 6));

	if ((arg1t == REG8_T)&&(arg1v == 6)) {
		arg1t = (isddfd)? IDXREF_T : REGREF_T;
		arg1v = 2;       /* HL, IX or IY */
	}

	if ((arg2t == REG8_T)&&(arg2v == 6)) {
		arg2t = (isddfd)? IDXREF_T : REGREF_T;
		arg2v = 2;       /* HL, IX or IY */
	}

	if (((arg1t == REG8_T)&&(arg2t == IDXREF_T))
	    ||((arg2t == REG8_T)&&(arg1t == IDXREF_T))) {
		reg8s_ = reg8s;
	}

	if (needfetch[arg1t] == 1) fprintf(fp, "\tint f1;\n");
	if (needfetch[arg1t] == 2) fprintf(fp, "\tint f1;\n");
	if (needfetch[arg2t] == 1) fprintf(fp, "\tint f2;\n");
	if (needfetch[arg2t] == 2) fprintf(fp, "\tint f2;\n");
	if (needfetch[arg1t] == 1) fprintf(fp, "\tfetchPC(f1);\n");
	if (needfetch[arg1t] == 2) fprintf(fp, "\tfetch2PC(f1);\n");
	if (needfetch[arg2t] == 1) fprintf(fp, "\tfetchPC(f2);\n");
	if (needfetch[arg2t] == 2) fprintf(fp, "\tfetch2PC(f2);\n");

	fprintf(fp, "\t%s%s%s%s(", op, (withld)? "_LD" : "", 
		midx[arg1t], midx[arg2t]);

	print_arg(fp, arg1t, arg1v, reg8s_, reg16s_, "f1");
	if (arg2t != NONE_T) {
		fprintf(fp, ",");
		print_arg(fp, arg2t, arg2v, reg8s_, reg16s_, "f2");
	}
	if (withld) {
		fprintf(fp, ",");
		print_arg(fp, REG8_T, ldarg, reg8s_, reg16s_, NULL);
	}

	fprintf(fp, ");\n");

	fprintf(fp, "#ifdef TRACE\n");
	fprintf(fp, "\tif (z80.trace) {\n");
	if (withld) {
		fprintf(fp, "\t\tprintf(\"LD\\t%s, \");\n", reg8s_[ldarg]);
	}
	fprintf(fp, "\t\tprintf(\"%s", op);
	if (arg1t != NONE_T) {
		fprintf(fp, "\\t");
		print_arg_for_code(fp, arg1t, arg1v, reg8s_, reg16s_);
		if (arg2t != NONE_T) {
			fprintf(fp, ", ");
			print_arg_for_code(fp, arg2t, arg2v, reg8s_, reg16s_);
		}
	}
	fprintf(fp, "\\n\"%s%s);\n",
		(needfetch[arg1t])? ", f1" : "",
		(needfetch[arg2t])? ", f2" : "" );
	fprintf(fp, "\t}\n");
	fprintf(fp, "#endif /* TRACE */\n");

	return 1;
}

static int print_func_body(FILE *fp, char *op, int arg1t, int arg1v,
			   int arg2t, int arg2v, int ddfd)
{
	return print_func_body_with_ld(fp, op, arg1t, arg1v, arg2t, arg2v, ddfd, -1);
}

static int print_func_body_noarg(FILE *fp, char *op)
{
	return print_func_body_with_ld(fp, op, NONE_T, 0, NONE_T, 0, 0, -1);
}

static int print_nop_func(FILE *fp, char *suffix, int i, int clockadd)
{
	print_func_header(fp, suffix, i);
	print_func_footer(fp, clockadd);
	return 1;
}

static int print_func_body_pushback(FILE *fp)
{
	fprintf(fp, "\tdecReg(PC);");
	return 1;
}


static int create_xxcbfile(int ddfd)
{
	FILE *fp;
	char *suffix, *filename;
	char *rotops[] = {"RLC", "RRC", "RL", "RR", "SLA", "SRA", "SLL", "SRL"};
	char *bitops[] = {"", "BIT", "RES", "SET"};
	int isddfd = (ddfd > 0);
	int i;

	suffix = (ddfd == 0)? SUFFIX_CB : ((ddfd == 1)? SUFFIX_DDCB : SUFFIX_FDCB);
	filename = (ddfd == 0)? FILE_CB : ((ddfd == 1)? FILE_DDCB : FILE_FDCB);

	fp = open_srcfile(filename, suffix);

	for (i = 0; i <= 0xff; i++) {
		int clockadd = 0;

		print_func_header(fp, suffix, i);

		if (i <= 0x3f) {
			print_func_body(fp, rotops[m3bit(i)], REG8_T, i&0x07, NONE_T, 0, ddfd);
			if ((i & 0x07) == 0x06) {
				clockadd = (isddfd)? 23 : 15;
			} else {
				clockadd = (isddfd)? 23 : 8;
			}
		} else if (i <= 0x7f) {
			if (isddfd) {
				print_func_body(fp, "BIT", CONST_T, m3bit(i), REG8_T, 6, ddfd);
			} else {
				print_func_body(fp, "BIT", CONST_T, m3bit(i), REG8_T, d3bit(i), ddfd);
			}
			if ((i & 0x07) == 0x06) {
				clockadd = (isddfd)? 20 : 12;
			} else {
				clockadd = (isddfd)? 20 : 8;
			}
		} else {
			if (isddfd) {
				print_func_body_with_ld(fp, bitops[u2bit(i)], CONST_T, m3bit(i), REG8_T, 6, ddfd, d3bit(i));
			} else {
				print_func_body(fp, bitops[u2bit(i)], CONST_T, m3bit(i), REG8_T, d3bit(i), ddfd);
			}
			if ((i & 0x07) == 0x06) {
				clockadd = (isddfd)? 23 : 15;
			} else {
				clockadd = (isddfd)? 23 : 8;
			}
		}
		print_func_footer(fp, clockadd);
	}

	close_srcfile(fp, suffix);
	return 1;
}

static int create_edfile()
{
	FILE *fp;
	char *suffix = SUFFIX_ED;
	int i;
	int imval[] = {0,0,1,2,0,0,1,2};
	char *edx7fops[] = {NULL, NULL, NULL, NULL, "RRD", "RLD", "NOP", "NOP"};
	int edx7fclks[] = {9,9,9,9,18,18,8,8};
	char *edabxops[] = {
		"LDI", "CPI", "INI", "OUTI", NULL, NULL, NULL, NULL,
		"LDD", "CPD", "IND", "OUTD", NULL, NULL, NULL, NULL,
		"LDIR", "CPIR", "INIR", "OTIR", NULL, NULL, NULL, NULL,
		"LDDR", "CPDR", "INDR", "OTDR", NULL, NULL, NULL, NULL,
	};

	fp = open_srcfile(FILE_ED, suffix);

	for (i = 0x00; i <= 0x3f; i++) {
		print_nop_func(fp, suffix, i, 8);
	}

	for (i = 0x40; i <= 0x7f; i++) {
		int clockadd = 0;
		print_func_header(fp, suffix, i);
		switch (d4bit(i)) {
		case 0x00: case 0x08:
			if (m3bit(i) != 6) {
				print_func_body(fp, "IN", REG8_T, m3bit(i), REGREF_T, 0, 0);
			} else {
				print_func_body(fp, "IN", REGREF_T, 0, NONE_T, 0, 0);
			}
			clockadd = 12;
			break;
		case 0x01: case 0x09:
			if (m3bit(i) != 6) {
				print_func_body(fp, "OUT", REGREF_T, 0, REG8_T, m3bit(i), 0);
			} else {
				print_func_body(fp, "OUT", REGREF_T, 0, CONST_T, 0, 0);
			}
			clockadd = 12;
			break;
		case 0x02:
			print_func_body(fp, "SBC", REG16_T, 2, REG16_T, m2bit(i), 0);
			clockadd = 15;
			break;
		case 0x0a:
			print_func_body(fp, "ADC", REG16_T, 2, REG16_T, m2bit(i), 0);
			clockadd = 15;
			break;
		case 0x03:
			print_func_body(fp, "LD", VALREF_T, 0, REG16_T, m2bit(i), 0);
			clockadd = 20;
			break;
		case 0x0b:
			print_func_body(fp, "LD", REG16_T, m2bit(i), VALREF_T, 0, 0);
			clockadd = 20;
			break;
		case 0x04: case 0x0c:
			print_func_body_noarg(fp, "NEG");
			clockadd = 8;
			break;
		case 0x05: case 0x0d:
			if (i == 0x4d) print_func_body_noarg(fp, "RETI");
			else print_func_body_noarg(fp, "RETN");
			clockadd = 14;
			break;
		case 0x06: case 0x0e:
			print_func_body(fp, "IM", CONST_T, imval[m3bit(i)], NONE_T, 0, 0);
			clockadd = 8;
			break;
		case 0x07: case 0x0f:
			switch (m3bit(i)) {
			case 0:
				print_func_body(fp, "LD", REG8_T, 9, REG8_T, 7, 0);
				break;
			case 1:
				print_func_body(fp, "LD", REG8_T, 8, REG8_T, 7, 0);
				break;
			case 2:
				print_func_body(fp, "LD", REG8_T, 7, REGRI_T, 9, 0);
				break;
			case 3:
				print_func_body(fp, "LD", REG8_T, 7, REGRI_T, 8, 0);
				break;
			default:
				print_func_body_noarg(fp, edx7fops[m3bit(i)]);
				break;
			}
			clockadd = edx7fclks[m3bit(i)];
			break;
		}
		print_func_footer(fp, clockadd);
	}

	for (i = 0x80; i <= 0x9f; i++) {
		print_nop_func(fp, suffix, i, 8);
	}

	for (i = 0xa0; i <= 0xbf; i++) {
		if (edabxops[i & 0x1f] == NULL) {
			print_nop_func(fp, suffix, i, 8);
		} else {
			print_func_header(fp, suffix, i);
			print_func_body_noarg(fp, edabxops[i & 0x1f]);
			if (i <= 0xaf) {
				print_func_footer(fp, 16);
			} else {
				print_func_footer_con(fp, "bitFlg(PV)", 21, 16);
			}
		}
	}

	for (i = 0xc0; i <= 0xff; i++) {
		switch(i) {
		case 0xcb: case 0xed: case 0xdd: case 0xfd:
			print_func_header(fp, suffix, i);
			print_func_body_pushback(fp);
			print_func_footer(fp, 4);
			break;
		default:
			print_nop_func(fp, suffix, i, 8);
			break;
		}
	}

	close_srcfile(fp, suffix);
	return 1;
}

static int create_xxfile(int ddfd)
{
	FILE *fp;
	char *suffix, *filename;
	int isddfd = (ddfd > 0);
	int i;
	char *xx07_3fops[] = {"RLCA", "RRCA", "RLA", "RRA",
			      "DAA", "CPL", "SCF", "CCF"};
	char *arops[] = {"ADD", "ADC", "SUB", "SBC",
			 "AND", "XOR", "OR", "CP"};
	int arops_reg[] = {1, 1, 0, 1, 0, 0, 0, 0};

	suffix = (ddfd == 0)? SUFFIX_XX : ((ddfd == 1)? SUFFIX_DD : SUFFIX_FD);
	filename = (ddfd == 0)? FILE_XX : ((ddfd == 1)? FILE_DD : FILE_FD);

	fp = open_srcfile(filename, suffix);

	for (i = 0x00; i <= 0x3f; i++) {
		int clockadd = 0;

		if ((i & 0x07) == 0x00) {
			print_func_header(fp, suffix, i);
			switch (m3bit(i)) {
			case 0:
				print_func_body_noarg(fp, "NOP");
				print_func_footer(fp, 4);
				break;
			case 1:
				print_func_body(fp, "EX", REG16_T, 4, REG16_T, 5, 0);
				print_func_footer(fp, 4);
				break;
			case 2:
				print_func_body(fp, "DJNZ", VALUE8_T, 4, NONE_T, 0, 0);
				print_func_footer_con(fp, "readReg(B)!=0", 13, 8);
				break;
			case 3:
				print_func_body(fp, "JR", VALUE8_T, 0, NONE_T, 0, 0);
				print_func_footer(fp, 12);
				break;
			default:
				print_func_body(fp, "JR", FLAG_T, m3bit(i)-4, VALUE8_T, 0, 0);
				print_func_footer_con(fp, flagconds[m3bit(i)-4], 12, 7);
				break;
			}
			continue;
		}

		print_func_header(fp, suffix, i);
		switch (i & 0x0f) {
		case 0x01:
			print_func_body(fp, "LD", REG16_T, m2bit(i), VALUE16_T, 0, ddfd);
			clockadd = 10;
			break;
		case 0x09:
			print_func_body(fp, "ADD", REG16_T, 2, REG16_T, m2bit(i), ddfd);
			clockadd = 11;
			break;
		case 0x02:
			switch(m2bit(i)) {
			case 0: case 1:
				print_func_body(fp, "LD", REGREF_T, m2bit(i), REG8_T, 7, ddfd);
				clockadd = 7;
				break;
			case 2:
				print_func_body(fp, "LD", VALREF_T, 0, REG16_T, 2, ddfd);
				clockadd = 16;
				break;
			case 3:
				print_func_body(fp, "LD", VALREF_T, 0, REG8_T, 7, ddfd);
				clockadd = 13;
				break;
			}
			break;
		case 0x0a:
			switch(m2bit(i)) {
			case 0: case 1:
				print_func_body(fp, "LD", REG8_T, 7, REGREF_T, m2bit(i), ddfd);
				break;
			case 2:
				print_func_body(fp, "LD", REG16_T, 2, VALREF_T, 0, ddfd);
				clockadd = 16;
				break;
			case 3:
				print_func_body(fp, "LD", REG8_T, 7, VALREF_T, 0, ddfd);
				clockadd = 13;
				break;
			}
			break;
		case 0x03:
			print_func_body(fp, "INC", REG16_T, m2bit(i), NONE_T, 0, ddfd);
			clockadd = 6;
			break;
		case 0x0b:
			print_func_body(fp, "DEC", REG16_T, m2bit(i), NONE_T, 0, ddfd);
			clockadd = 6;
			break;
		case 0x04: case 0x0c:
			print_func_body(fp, "INC", REG8_T, m3bit(i), NONE_T, 0, ddfd);
			clockadd = (m3bit(i)!=6)? 4 : ((isddfd)? 19 : 11);
			break;
		case 0x05: case 0x0d:
			print_func_body(fp, "DEC", REG8_T, m3bit(i), NONE_T, 0, ddfd);
			clockadd = (m3bit(i)!=6)? 4 : ((isddfd)? 19 : 11);
			break;
		case 0x06: case 0x0e:
			print_func_body(fp, "LD", REG8_T, m3bit(i), VALUE8_T, 0, ddfd);
			clockadd = (m3bit(i)!=6)? 7 : ((isddfd)? 15 : 10);
			break;
		case 0x07: case 0x0f:
			print_func_body_noarg(fp, xx07_3fops[m3bit(i)]);
			clockadd = 4;
			break;
		}
		print_func_footer(fp, clockadd);
	}

	for (i = 0x40; i <= 0x7f; i++) {
		int clockadd = 4;
		print_func_header(fp, suffix, i);
		if (i != 0x76) {
			print_func_body(fp, "LD", REG8_T, m3bit(i), REG8_T, d3bit(i), ddfd);
			if ((m3bit(i)==6)||(d3bit(i)==6)) clockadd += (isddfd)? 11 : 3;
		} else {
			print_func_body_noarg(fp, "HALT");
		}
		print_func_footer(fp, clockadd);
	}

	for (i = 0x80; i <= 0xbf; i++) {
		int clockadd = 4;
		print_func_header(fp, suffix, i);
		if (arops_reg[m3bit(i)]) {
			print_func_body(fp, arops[m3bit(i)], REG8_T, 7, REG8_T, d3bit(i), ddfd);
		} else {
			print_func_body(fp, arops[m3bit(i)], REG8_T, d3bit(i), NONE_T, 0, ddfd);
		}
		if (d3bit(i)==6) clockadd += (isddfd)? 11 : 3;
		print_func_footer(fp, clockadd);
	}

	for (i = 0xc0; i <= 0xff; i++) {
		print_func_header(fp, suffix, i);
		switch (d4bit(i)) {
		case 0x00: case 0x08:
			print_func_body(fp, "RET", FLAG_T, m3bit(i), NONE_T, 0, ddfd);
			print_func_footer_con(fp, flagconds[m3bit(i)], 11, 5);
			break;
		case 0x01:
			if (m2bit(i) != 3) {
				print_func_body(fp, "POP", REG16_T, m2bit(i), NONE_T, 0, ddfd);
			} else {
				print_func_body(fp, "POP", REG16_T, 4, NONE_T, 0, ddfd);
			}
			print_func_footer(fp, 10);
			break;
		case 0x09:
			switch(m2bit(i)) {
			case 0:
				print_func_body_noarg(fp, "RET");
				print_func_footer(fp, 10);
				break;
			case 1:
				print_func_body_noarg(fp, "EXX");
				print_func_footer(fp, 4);
				break;
			case 2:
				print_func_body(fp, "JP", REG16_T, 2, NONE_T, 0, ddfd);
				print_func_footer(fp, 4);
				break;
			case 3:
				print_func_body(fp, "LD", REG16_T, 3, REG16_T, 2, ddfd);
				print_func_footer(fp, 6);
				break;
			}
			break;
		case 0x02: case 0x0a:
			print_func_body(fp, "JP", FLAG_T, m3bit(i), VALUE16_T, 0, ddfd);
			print_func_footer(fp, 10);
			break;
		case 0x03: case 0x0b:
			switch (m3bit(i)) {
			case 0:
				print_func_body(fp, "JP", VALUE16_T, 0, NONE_T, 0, ddfd);
				print_func_footer(fp, 10);
				break;
			case 1:
				if (ddfd == 0) {
					print_func_footer_str(fp, "exec_instruction(GROUP_CB)");
				} else if (ddfd == 1) {
					print_func_footer_str(fp, "exec_instruction2(GROUP_DDCB)-4");
				} else if (ddfd == 2) {
					print_func_footer_str(fp, "exec_instruction2(GROUP_FDCB)-4");
				}
				break;
			case 2:
				print_func_body(fp, "OUT", VAL8REF_T, 0, REG8_T, 7, ddfd);
				print_func_footer(fp, 11);
				break;
			case 3:
				print_func_body(fp, "IN", REG8_T, 7, VAL8REF_T, 0, ddfd);
				print_func_footer(fp, 11);
				break;
			case 4:
				print_func_body(fp, "EX", REGREF_T, 3, REG16_T, 2, ddfd);
				print_func_footer(fp, 19);
				break;
			case 5:
				print_func_body(fp, "EX", REG16_T, 1, REG16_T, 2, ddfd);
				print_func_footer(fp, 4);
				break;
			case 6:
				print_func_body_noarg(fp, "DI");
				print_func_footer(fp, 4);
				break;
			case 7:
				print_func_body_noarg(fp, "EI");
				print_func_footer(fp, 4);
				break;
			}
			break;
		case 0x04: case 0x0c:
			print_func_body(fp, "CALL", FLAG_T, m3bit(i), VALUE16_T, 0, ddfd);
			print_func_footer_con(fp, flagconds[m3bit(i)], 17, 10);
			break;
		case 0x05:
			if (m2bit(i) != 3) {
				print_func_body(fp, "PUSH", REG16_T, m2bit(i), NONE_T, 0, ddfd);
			} else {
				print_func_body(fp, "PUSH", REG16_T, 4, NONE_T, 0, ddfd);
			}
			print_func_footer(fp, 10);
			break;
		case 0x0d:
			switch(m2bit(i)) {
			case 0:
				print_func_body(fp, "CALL", VALUE16_T, 0, NONE_T, 0, ddfd);
				print_func_footer(fp, 17);
				break;
			case 1:
				if (isddfd) {
					print_func_body_pushback(fp);
					print_func_footer(fp, 0);
				} else {
					print_func_footer_str(fp, "exec_instruction(GROUP_DD)+4");
				}
				break;
			case 2:
				if (isddfd) {
					print_func_body_pushback(fp);
					print_func_footer(fp, 0);
				} else {
					print_func_footer_str(fp, "exec_instruction(GROUP_ED)");
				}
				break;
			case 3:
				if (isddfd) {
					print_func_body_pushback(fp);
					print_func_footer(fp, 0);
				} else {
					print_func_footer_str(fp, "exec_instruction(GROUP_FD)+4");
				}
				break;
			}
			break;
		case 0x06: case 0x0e:
			if (arops_reg[m3bit(i)]) {
				print_func_body(fp, arops[m3bit(i)], REG8_T, 7, VALUE8_T, 0, ddfd);
			} else {
				print_func_body(fp, arops[m3bit(i)], VALUE8_T, 0, NONE_T, 0, ddfd);
			}
			print_func_footer(fp, 7);
			break;
		case 0x07: case 0x0f:
			print_func_body(fp, "RST", CONST_T, i-0xc7, NONE_T, 0, ddfd);
			print_func_footer(fp, 11);
			break;
		}
	}

	close_srcfile(fp, suffix);
	return 1;
}

int main()
{
	create_xxcbfile(0);  /* CB */
	create_xxcbfile(1);  /* DDCB */
	create_xxcbfile(2);  /* FDCB */
	create_edfile();     /* ED */
	create_xxfile(0);    /* xx */
	create_xxfile(1);    /* DD */
	create_xxfile(2);    /* FD */
	return 0;
}
