#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "tape.h"

TAPE *tropen(char *name)
{
    FILE *fp;
    TAPE *tp;
    char head[8];

    if ((fp = fopen(name, "r")) == NULL) return NULL;
    fread(head, 1, 8, fp);
    if (memcmp(head, "mz20", 4)) {
	fclose(fp);
	return NULL;
    }

    if (head[4] != 0x00) {
	fclose(fp);
	return NULL;
    }

    switch(head[5]) {
    case RUNLENGTH:
    case FILE2000B:
	break;
    default:
	fclose(fp);
	return NULL;
    }

    tp = (TAPE *)malloc(sizeof(TAPE));

    tp->fp = fp;
    tp->format = head[5];
    tp->state = 0;
    tp->hl = 0;
    tp->clock = 0;

    switch (tp->format) {
    case FILE2000B:
	tp->data = 0;
	tp->nbit = 0;
	tp->sl = 0;
	tp->send = 0;
	tp->shortclk = 666;
	tp->longclk = 1334;
	break;
    }

    return tp;
}

TAPE *twopen(char *name, int format)
{
    FILE *fp;
    TAPE *tp;
    char head[8];

    switch(format) {
    case RUNLENGTH:
    case FILE2000B:
	break;
    default:
	return NULL;
    }

    if ((fp = fopen(name, "w")) == NULL) return NULL;
    memcpy(head, "mz20", 4);
    head[4] = 0x00;
    head[5] = format;
    head[6] = 0x00;
    head[7] = 0x00;

    fwrite(head, 1, 8, fp);

    tp = (TAPE *)malloc(sizeof(TAPE));

    tp->fp = fp;
    tp->format = format;
    tp->state = 0;
    tp->hl = 0;
    tp->clock = 0;

    switch (format) {
    case FILE2000B:
	tp->data = 0;
	tp->nbit = 0;
	tp->sl = 0;
	tp->send = 0;
	tp->shortclk = 1000;
	tp->longclk = 1000;
	tp->mstate = 0;
	tp->size = 0;
	break;
    }
    return tp;
}

int tclose(TAPE *tp)
{
    fclose(tp->fp);
    free(tp);
    return 1;
}

int tscan(TAPE *tp)
{
    return(tp->hl);
}

static int getword(FILE *fp)
{
    int x, y;

    x = getc(fp);
    if (x == EOF) return -1;
    y = getc(fp);
    if (y == EOF) return -1;
    return (x * 256 + y + 1);
}

static long getclk(FILE *fp)
{
    int x, y;

    x = getword(fp);
    if (x == -1) return -1;
    if (x != 0xffff) return x + 1;

    x = getword(fp);
    if (x == -1) return -1;
    y = getword(fp);
    if (y == -1) return -1;
    return x * 65536 + y + 1;
}

int tskip(TAPE *tp, long clock)
{
    if (tp->format == RUNLENGTH) {
	int x;

	while (tp->clock < clock) {
	    switch (tp->state) {
	    case 0:  /* blank before data */
		tp->state++;
		tp->clock += 16000000;  /* 4sec */
		tp->hl = 0;
		break;
	    case 1:  /* data */
		x = getclk(tp->fp);
		if (x == -1) {
		    tp->state++;
		    tp->clock += 16000000; /* 4sec */
		    tp->hl = 0;
		    break;
		}
		tp->hl = (tp->hl == 0)? 1 : 0;
		tp->clock += x;
		break;
	    case 2:  /* blank after data */
		tp->hl = -1;
		tp->state++;
		break;
	    case 3:
		tp->clock += 16000000;
		break;
	    }
	}
	tp->clock -= clock;
    } else if (tp->format == FILE2000B) {
	while (tp->clock < clock) {
	    switch (tp->state) {
	    case 0:  /* blank before data (4sec blank) */
	    case 9: /* blank between info and data (4sec blank)*/
		tp->state++;
		tp->clock += 16000000;  /* 4sec */
		tp->hl = 0;
		tp->data = (tp->state == 1)? 0x2710 : 0x2AF8;
		tp->sl = 0;
		break;
	    case 1:  /* inf block -- gap1 */
	    case 2:  /* inf block -- gap2 */
	    case 3:  /* inf block -- gap3 */
	    case 4:  /* inf block -- gap4 */
	    case 6:
	    case 8:
	    case 10:  /* data block -- gap1 */
	    case 11:  /* data block -- gap2 */
	    case 12:  /* data block -- gap3 */
	    case 13:  /* data block -- gap4 */
	    case 15:
	    case 17:
		tp->clock += (tp->sl == 0)? tp->shortclk : tp->longclk;
		if (tp->hl == 0) {
		    tp->hl = 1;
		} else {
		    tp->hl = 0;
		    if (--(tp->data) == 0) {
			switch (tp->state) {
			case 1:
			    tp->data = 40; tp->sl = 1;
			    break;
			case 2:
			    tp->data = 40; tp->sl = 0;
			    break;
			case 3:
			case 12:
			    tp->data = 1; tp->sl = 1;
			    break;
			case 4:
			case 6:
			case 13:
			case 15:
			    tp->data = getc(tp->fp);
			    tp->nbit = 8;
			    tp->sl = 1;
			    break;
			case 8:
			    break;
			case 10:
			    tp->data = 20; tp->sl = 1;
			    break;
			case 11:
			    tp->data = 20; tp->sl = 0;
			    break;
			case 17:
			    tp->clock += 16000000; /* 4sec */
			    break;
			}
			tp->state++;
		    }
		}
		break;
	    case 5: /* inf block -- data */
	    case 7: /* inf block -- data 2nd */
	    case 14: /* data block -- data */
	    case 16: /* data block -- data 2nd */
		tp->clock += (tp->sl == 0)? tp->shortclk : tp->longclk;
		if (tp->hl == 0) {
		    tp->hl = 1;
		} else {
		    tp->hl = 0;
		    if (--tp->nbit < 0) {
			tp->send++;
			if (tp->send == 0x82) {
			    tp->data = 1; tp->sl = 1;
			    tp->state++;
			    fseek(tp->fp, 0x08, SEEK_SET);
			    break;
			} else if (tp->send == 0x104) {
			    tp->data = 1; tp->sl = 1;
			    tp->state++;
			    break;
			}
			tp->data = getc(tp->fp);
			if (tp->data == EOF) {
			    if (tp->state == 16) {
				tp->data = 1; tp->sl = 1;
				tp->state++;
				break;
			    } else {
				tp->data = 1; tp->sl = 1;
				tp->state++;
				fseek(tp->fp, 0x8A, SEEK_SET);
				break;
			    }
			}
			tp->nbit = 8;
			tp->sl = 1;
		    } else {
			tp->sl = ((tp->data) >> tp->nbit) & 0x01;
		    }
		}
		break;
	    case 18:  /* blank after data */
		tp->hl = -1;
		tp->state++;
		break;
	    case 19:
		tp->clock += 16000000;
		break;
	    }
	}
	tp->clock -= clock;
    }
    return tp->hl;
}

static int putword(int val, FILE *fp)
{
    putc((val >> 8) & 0xff, fp);
    putc(val & 0xff, fp);

    return 1;
}

static int putclk(long clock, FILE *fp)
{
    clock--;
    if (clock < 0xffff) {
	putword(clock, fp);
    } else {
	putword(0xffff, fp);
	putword((clock >> 16) & 0xffff, fp);
	putword(clock & 0xffff, fp);
    }
    return 1;
}

int prev = 0;

int twrite(TAPE *tp, int bit, long clock)
{
    bit = (bit == 0)? 0 : 1;
    if (tp->format == RUNLENGTH) {
	switch (tp->state) {
	case 0:
	    if (bit != 0) {
		tp->hl = 1;
		tp->clock = clock;
		tp->state++;
	    }
	    break;
	case 1:
	    if (bit == tp->hl) {
		tp->clock += clock;
	    } else {
		putclk(tp->clock, tp->fp);
		tp->hl = bit;
		tp->clock = clock;
	    }
	    break;
	}
    } else if (tp->format == FILE2000B) {
	switch (tp->mstate) {
	case 0: /* seeking low (for seeking edge) */
	    if (bit == 0) tp->mstate++;
	    break;
	case 1: /* seeking high (for seeking edge) */
	    if (bit == 0) break;
	    tp->mstate++;
	    tp->clock = tp->shortclk;
	    /* not break */
	case 2: /* check short or long region */
	    if (tp->clock > clock) {
		tp->clock -= clock;
	    } else {
		switch (tp->state) {
		case 0:
		    if (bit == 1) {
			tp->data++;
			if (tp->data == 40) {
			    tp->state++;
			    tp->data = 0;
			}
		    } else {
			tp->data = 0;
		    }
		    break;
		case 1:
		    if (bit == 0) {
			tp->data++;
			if (tp->data == 40) {
			    tp->state++;
			    tp->data = 0;
			}
		    } else {
			tp->state--;
			tp->data = 0;
		    }
		    break;
		case 2:
		    tp->state++;
		    break;
		case 3:
		    if (bit == 1) {
			tp->state++;
			tp->data = 0;
		    }
		    break;
		case 4:
		    if (++(tp->nbit) == 9) {
			putc(tp->data & 0xff, tp->fp);
			if (tp->send == 0x12) tp->size = tp->data;
			if (tp->send == 0x13) tp->size += tp->data * 256;
			tp->nbit = 0;
			tp->data = 0;
			if (++(tp->send) == 0x82) {
			    tp->state++;
			}
		    } else {
			tp->data = tp->data << 1;
			tp->data |= (bit & 0x01);
		    }
		    break;
		case 5:
		    if (bit == 1) {
			tp->data++;
			if (tp->data == 20) {
			    tp->state++;
			    tp->data = 0;
			}
		    } else {
			tp->data = 0;
		    }
		    break;
		case 6:
		    if (bit == 0) {
			tp->data++;
			if (tp->data == 20) {
			    tp->state++;
			    tp->data = 0;
			}
		    } else {
			tp->state--;
			tp->data = 0;
		    }
		    break;
		case 7:
		    tp->state++;
		    break;
		case 8:
		    if (bit == 1) {
			tp->state++;
			tp->data = 0;
			tp->send = 0;
		    }
		    break;
		case 9:
		    if (++(tp->nbit) == 9) {
			putc(tp->data & 0xff, tp->fp);
			tp->nbit = 0;
			tp->data = 0;
			if (++(tp->send) == tp->size + 2) {
			    tp->state++;
			}
		    } else {
			tp->data = tp->data << 1;
			tp->data |= (bit & 0x01);
		    }
		    break;
		case 10:
		    break;
		}
		tp->mstate = (bit == 0)? 1 : 0;
	    }
	}
    }
    return 1;
}
