#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include <ctype.h>
#include <pthread.h>
#include "mz2000.h"

static int screen_setcolor(MZ2000 *mz2000);

static MZFont *font_read(char *fontfile)
{
	MZFont *font;
	FILE *fp;
	int i, j, k;

	if ((font = (MZFont *)malloc(sizeof(MZFont))) == NULL) {
		fprintf(stderr, "No enough memory\n");
		return NULL;
	}

	if ((fp = fopen(fontfile, "r")) == NULL) {
		fprintf(stderr, "Can't find font file: %s\n", fontfile);
		free(font);
		return NULL;
	}
	for (i = 0; i < 256; i++) {
		for (j = 0; j < 8; j++) {
			for (k = 0; k < 8; k++) {
				while (1) {
					int c = getc(fp);
					if (isspace(c)) continue;
					else if (c == EOF) {
						fprintf(stderr, "Font file format error (%d,%d)\n",
							i, j);
						free(font);
						return NULL;
					}
					else if (c == '#') {
						while ((c != '\n')&&(c != EOF)) c = getc(fp);
					} else if (c == 'X') {
						font->chr40[i][j*16+k*2] = 8;
						font->chr40[i][j*16+k*2+1] = 8;
						font->chr80[i][j*8+k] = 8;
						break;
					} else if (c == '.') {
						font->chr40[i][j*16+k*2] = 0;
						font->chr40[i][j*16+k*2+1] = 0;
						font->chr80[i][j*8+k] = 0;
						break;
					} else {
						fprintf(stderr,
							"Font file unexpected character: %c\n",c);
						free(font);
						return NULL;
					}
				}
			}
		}
	}
	return font;
}

int init_screen(MZ2000 *mz2000, char *fontfile)
{
	MZScreen *scr;
	int i;

	if ((scr = (MZScreen *)calloc(sizeof(MZScreen),1)) == NULL) {
		fprintf(stderr, "No enough memory\n");
		return 0;
	}
	mz2000->screen = scr;

	/* font read */
	if ((scr->font = font_read(fontfile)) == NULL) {
		free(scr);
		return 0;
	}

	mzgtk_init_window(mz2000);
	scr->gc = gdk_gc_new(scr->screen_widget->window);
	gdk_rgb_init();

	/* region */
	scr->region.valid = 1;

	/* output registers */
	scr->F4 = 0;
	scr->F5 = 7;
	scr->F6 = 0;

	/* screen default */
	scr->issize40 = 1;
	scr->isgreen = 1;

	/* Colors */
	for (i = 0; i < 8; i++) {
		guint32 p = 0;
		if (i & 0x01) p |= 0xff;
		if (i & 0x02) p |= 0xff0000;
		if (i & 0x04) p |= 0xff00;
		scr->pixel[i] = p;
	}
	screen_setcolor(mz2000);

	pthread_mutex_init(&(scr->reg_mutex), NULL);

	return 1;
}

int screen_portF4(MZ2000 *mz2000, unsigned char val)
{
	mz2000->screen->F4 = val & 7;
	screen_setcolor(mz2000);
	return screen_region_add_area(mz2000, 0, 0, 640, 400);
}

int screen_portF5(MZ2000 *mz2000, unsigned char val)
{
	mz2000->screen->F5 = val & 15;
	screen_setcolor(mz2000);
	return screen_region_add_area(mz2000, 0, 0, 640, 400);
}

int screen_portF6(MZ2000 *mz2000, unsigned char val)
{
	mz2000->screen->F6 = val & 15;
	screen_setcolor(mz2000);
	return screen_region_add_area(mz2000, 0, 0, 640, 400);
}

int screen_color(MZ2000 *mz2000)
{
	mz2000->screen->isgreen = 0;
	screen_setcolor(mz2000);
	screen_region_add_area(mz2000, 0, 0, 640, 400);
	return 1;
}

int screen_green(MZ2000 *mz2000)
{
	mz2000->screen->isgreen = 1;
	screen_setcolor(mz2000);
	screen_region_add_area(mz2000, 0, 0, 640, 400);
	return 1;
}

int screen_size40(MZ2000 *mz2000)
{
	mz2000->screen->issize40 = 1;
	return screen_region_add_area(mz2000, 0, 0, 640, 400);
}

int screen_size80(MZ2000 *mz2000)
{
	mz2000->screen->issize40 = 0;
	return screen_region_add_area(mz2000, 0, 0, 640, 400);
}

static int screen_setcolor(MZ2000 *mz2000)
{
	int i, c;
	MZScreen *scr = mz2000->screen;

	if (scr->isgreen) {
		int gc = (scr->F6 & 8)? 0 : (scr->F6 & 7);
		for (i = 0; i < 16; i++) {
			if ((i >= 8)||((gc & i) != 0)) {
				c = 4;
			} else {
				c = 0;
			}
			scr->colors[i] = scr->pixel[c];
		}
	} else {
		scr->colors[0] = scr->pixel[scr->F4];
		for (i = 1; i < 16; i++) {
			int go = i & scr->F6 & 7, co = (i >= 8);
			int pg = scr->F5 & 8;

			if (go && co) {
				c = (pg)? i : scr->F5;
			} else if (!go && !co) {
				c = scr->F4;
			} else if (go) {
				c = i;
			} else if (co) {
				c = scr->F5;
			}
			scr->colors[i] = scr->pixel[c];
		}
	}
	if (scr->cmap != NULL) {
		gdk_rgb_cmap_free(scr->cmap);
	}
	scr->cmap = gdk_rgb_cmap_new(scr->colors, 16);

	return 1;
}

static int screen_ram2buf(MZ2000 *mz2000, int x, int y, int width, int height)
{
	MZScreen *scr = mz2000->screen;
	int i, j, k;
	int chx1, chy1, chx2, chy2;
	int grx1, gry1, grx2, gry2;

	if (scr->issize40) {
		chx1 = x / 16;
		chy1 = y / 16;
		chx2 = (x + width + 15) / 16;
		chy2 = (y + height + 15) / 16;

		for (i = chy1; i < chy2; i++) {
			for (j = chx1; j < chx2; j++) {
				u8 c = mz2000->vram[i*40+j];
				for (k = 0; k < 8; k++) {
					memcpy(&(scr->scrbuf[i*640*16 + k*640*2 + j*16]),
					       &(scr->font->chr40[c][k*16]), 16);
					memcpy(&(scr->scrbuf[i*640*16 + k*640*2 + j*16 + 640]),
					       &(scr->font->chr40[c][k*16]), 16);
				}
			}
		}
	} else {
		chx1 = x / 8;
		chy1 = y / 16;
		chx2 = (x + width + 7) / 8;
		chy2 = (y + height + 15) / 16;

		for (i = chy1; i < chy2; i++) {
			for (j = chx1; j < chx2; j++) {
				u8 c = mz2000->vram[i*80+j];
				for (k = 0; k < 8; k++) {
					memcpy(&(scr->scrbuf[i*640*16 + k*640*2 + j*8]),
					       &(scr->font->chr80[c][k*8]), 8);
					memcpy(&(scr->scrbuf[i*640*16 + k*640*2 + j*8 + 640]),
					       &(scr->font->chr80[c][k*8]), 8);
				}
			}
		}
	}

	grx1 = x / 8;
	gry1 = y / 2;
	grx2 = (x + width + 7) / 8;
	gry2 = (y + height + 1) / 2;
	for (i = gry1; i < gry2; i++) {
		for (j = grx1; j < grx2; j++) {
			for (k = 0; k < 8; k++) {
				u8 c;
				c = ((mz2000->gram1[i*80+j] >> k) & 1)? 1 : 0;
				c += ((mz2000->gram2[i*80+j] >> k) & 1)? 2 : 0;
				c += ((mz2000->gram3[i*80+j] >> k) & 1)? 4 : 0;
				scr->scrbuf[i*640*2 + j*8  + k] += c;
				scr->scrbuf[i*640*2 + j*8  + k + 640] += c;
			}
		}
	}
	return 1;
}

int screen_redraw(MZ2000 *mz2000, int x, int y, int width, int height)
{
	MZScreen *scr = mz2000->screen;

	screen_ram2buf(mz2000, x, y, width, height);
	gdk_draw_indexed_image(scr->screen_widget->window, scr->gc,
			       x, y, width, height,
			       GDK_RGB_DITHER_NORMAL, scr->scrbuf + x + y*640,
			       640, scr->cmap);

	return 1;
}

int screen_redraw_all(MZ2000 *mz2000)
{
	MZScreen *scr = mz2000->screen;

	screen_redraw(mz2000, 0, 0, 640, 400);
	pthread_mutex_lock(&(scr->reg_mutex));
	scr->region.valid = 0;
	pthread_mutex_unlock(&(scr->reg_mutex));
	return 1;
}

int screen_redraw_region(MZ2000 *mz2000)
{
	int x, y, width, height;
	MZScreen *scr = mz2000->screen;

	pthread_mutex_lock(&(scr->reg_mutex));
	if (scr->region.valid) {
		x = scr->region.x1;
		y = scr->region.y1;
		width = scr->region.x2 - scr->region.x1;
		height = scr->region.y2 - scr->region.y1;
	} else {
		x = y = width = height = -1;
	}
	scr->region.valid = 0;
	pthread_mutex_unlock(&(scr->reg_mutex));
	if (x >= 0) screen_redraw(mz2000, x, y, width, height);
	return 1;
}

int screen_region_add_area(MZ2000 *mz2000, int x, int y, int width, int height)
{
	MZScreen *scr = mz2000->screen;

	pthread_mutex_lock(&(scr->reg_mutex));
	if (scr->region.valid) {
		if (scr->region.x1 > x) scr->region.x1 = x;
		if (scr->region.y1 > y) scr->region.y1 = y;
		if (scr->region.x2 < x + width) scr->region.x2 = x + width;
		if (scr->region.y2 < y + height) scr->region.y2 = y + height;
	} else {
		scr->region.x1 = x;
		scr->region.y1 = y;
		scr->region.x2 = x + width;
		scr->region.y2 = y + height;
		scr->region.valid = 1;
	}
	pthread_mutex_unlock(&(scr->reg_mutex));
	return 1;
}
