#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "tape.h"

/*
 *  tp->flags
 */
#define TAPE_WRITE_FLAG 0x01
#define TAPE_FILE_FLAG  0x02
#define TAPE_FINISH_FLAG  0x04
#define TAPE_ERROR_FLAG 0x08
#define TAPE_NOWRITE_FLAG 0x10
#define tformat(tp) (((tp)->flags)>>8)
#define tiswrite(tp) (((tp)->flags) & TAPE_WRITE_FLAG)
#define tisfile(tp) (((tp)->flags) & TAPE_FILE_FLAG)
#define tisfinish(tp) (((tp)->flags) & TAPE_FINISH_FLAG)
#define tiserror(tp) (((tp)->flags) & TAPE_ERROR_FLAG)
#define tnowrite(tp) (((tp)->flags) & TAPE_NOWRITE_FLAG)
#define tsetflags(tp,tformat,iswrite,isfile) \
	(((tp)->flags) = ((tformat)<<8)\
		| ((iswrite)? TAPE_WRITE_FLAG : 0)\
		| ((isfile)? TAPE_FILE_FLAG : 0))
#define tsetfinish(tp) (((tp)->flags) |= TAPE_FINISH_FLAG)
#define tresetfinish(tp) (((tp)->flags) &= (~TAPE_FINISH_FLAG))
#define tseterror(tp) ((((tp)->flags) |= TAPE_ERROR_FLAG),(tp->state = -1))
#define tsetnowrite(tp) ((((tp)->flags) |= TAPE_NOWRITE_FLAG))
#define tresetnowrite(tp) (((tp)->flags) &= (~TAPE_NOWRITE_FLAG))
#define validreadformat(format) ((format > 0)&&(format <= VALIDREADFORMATS))
#define validwriteformat(format) ((format > 0)&&(format <= VALIDWRITEFORMATS))

#define MAXHEADERSIZE 8

static int getbyte(TAPE *tp);
static int getword_be(TAPE *tp);
static int getword_le(TAPE *tp);
static z80clk getclk(TAPE *tp);
static int putbyte(int val, TAPE *tp);
static int putword_be(int val, TAPE *tp);
static int putword_le(int val, TAPE *tp);
static int putclk(z80clk clock, TAPE *tp);
static int count_nbits(int val);

static TBUFFER *tbopen(unsigned char *buffer, int bufsize);
static int tbclose(TBUFFER *bp);

static int tseek(TAPE *tp, int offset, int whence);
#define TSEEK_HEADER  1
#define TSEEK_MAINBLK 2
#define TSEEK_INFOBLK 3
#define TSEEK_DATABLK 4
static int tflush(TAPE *tp);

/*
 *  static int getbyte(TAPE *tp)
 *  reading 1byte data
 */
static int getbyte(TAPE *tp)
{
	int x;
	if (tisfile(tp)) {
		x = getc(tp->fp);
		if (x == EOF) return -1;
	} else {
		if (tp->bp->curoffset < tp->bp->bufsize) {
			x = tp->bp->buffer[tp->bp->curoffset++];
		} else {
			x = -1;
		}
	}
	return x;
}

/*
 *  static int getword_be(TAPE *tp)
 *  reading word(2bytes,big endian) data;
 */
static int getword_be(TAPE *tp)
{
	int x, y;

	if ((x = getbyte(tp)) == -1) return -1;
	if ((y = getbyte(tp)) == -1) return -1;
	return (x * 256 + y);
}

/*
 *  static int getword_le(TAPE *tp)
 *  reading a word(2bytes,little endian) data;
 */
static int getword_le(TAPE *tp)
{
	int x, y;

	if ((x = getbyte(tp)) == -1) return -1;
	if ((y = getbyte(tp)) == -1) return -1;
	return (y * 256 + x);
}

/*
 *  static int getclk(TAPE *tp)
 *  reading clock(1bytes or 2bytes) data;
 */
static z80clk getclk(TAPE *tp)
{
	int x, y;

	x = getword_be(tp);
	if (x == -1) return -1;
	if (x != 0xffff) return x + 1;

	x = getword_be(tp);
	if (x == -1) return -1;
	y = getword_be(tp);
	if (y == -1) return -1;
	return x * 65536 + y + 1;
}

/*
 *  static int putbyte(TAPE *tp)
 *  writing 1byte data
 */
static int putbyte(int val, TAPE *tp)
{
	int x;

	if (tnowrite(tp)) return 1;
	if (tisfile(tp)) {
		x = putc(val, tp->fp);
	} else {
		if (tp->bp->curoffset < tp->bp->bufsize) {
			tp->bp->buffer[tp->bp->curoffset++] = val;
			x = 1;
		} else {
			x = -1;
		}
	}
	return x;
}

/*
 *  static int putword_be(TAPE *tp)
 *  writing word(2bytes,big endian) data;
 */
static int putword_be(int val, TAPE *tp)
{
	if ((putbyte((val >> 8) & 0xff, tp)) == -1) return -1;
	if ((putbyte(val & 0xff, tp)) == -1) return -1;

	return 1;
}

/*
 *  static int putword_le(TAPE *tp)
 *  writing word(2bytes,little endian) data;
 */
static int putword_le(int val, TAPE *tp)
{
	if ((putbyte(val & 0xff, tp)) == -1) return -1;
	if ((putbyte((val >> 8) & 0xff, tp)) == -1) return -1;

	return 1;
}

/*
 *  static int putclk(z80clk clock, TAPE *tp)
 *  writing clock(1bytes or 2bytes) data;
 */
static int putclk(z80clk clock, TAPE *tp)
{
	clock--;
	if (clock < 0xffff) {
		if (putword_be(clock, tp) == -1) return -1;
	} else {
		if (putword_be(0xffff, tp) == -1) return -1;
		if (putword_be((clock >> 16) & 0xffff, tp) == -1) return -1;
		if (putword_be(clock & 0xffff, tp) == -1) return -1;
	}
	return 1;
}

/*
 * int count_nbits(int val)
 * count 1 bits in val
 */
int count_nbits(int val)
{
	int i, nbits = 0;
	for (i = 1; i < 0x100; i *= 2) {
		if (val & i) nbits++;
	}
	return nbits;
}

/*
 *  TBUFFER *tbopen(unsigned char *buffer, int bufsize)
 *  open new buffer handler
 */
static TBUFFER *tbopen(unsigned char *buffer, int bufsize)
{
	TBUFFER *bp;

	if ((bp = (TBUFFER *)malloc(sizeof(TBUFFER))) == NULL) return NULL;
	bp->buffer = buffer;
	bp->bufsize = bufsize;
	bp->curoffset = 0;
	return bp;
}

/*
 *  int tbclose(TBUFFER *bp)
 *  close new buffer handler
 */
static int tbclose(TBUFFER *bp)
{
	free(bp);
	return 0;
}

/*
 *  Functions for opening tape handle
 */
static TAPE *tpinit(int format, int iswrite, FILE *fp, TBUFFER *bp);
static int tscan_header(unsigned char *header);
static int tmake_header(unsigned char *header, int format);

/*
 * TAPE *tropen(char *name)
 *  Opening a named tape image file for reading.
 *  name : filename
 */
TAPE *tropen(char *name)
{
	FILE *fp;
	TAPE *tp;
	char header[MAXHEADERSIZE];
	int format;

	if ((fp = fopen(name, "r")) == NULL) return NULL;
	fread(header, 1, MAXHEADERSIZE, fp);

	format = tscan_header(header);
	if (format == 0) {
		fprintf(stderr, "Invalid read format\n");
		fclose(fp);
		return NULL;
	}
	tp = tpinit(format, 0, fp, NULL);
	if (tp == NULL) {
		fclose(fp);
		return NULL;
	}

	return tp;
}

/*
 * TAPE *tropen_mem(unsigned char *buffer, int bufsize)
 *  Opening a tape image written in memory for reading.
 *  buffer  : pointer to memory
 *  bufsize : size of available memory
 */
TAPE *tropen_mem(unsigned char *buffer, int bufsize)
{
	TBUFFER *bp;
	TAPE *tp;
	int format;

	format = tscan_header(buffer);
	if (format == 0) {
		fprintf(stderr, "Invalid read format\n");
		return NULL;
	}

	bp = tbopen(buffer, bufsize);
	if (bp == NULL) return NULL;
	tp = tpinit(format, 0, NULL, bp);
	if (tp == NULL) {
		tbclose(bp);
		return NULL;
	}
	
	return tp;
}

/*
 * TAPE *twopen(char *name, int format)
 *  Opening a new tape image file for writing.
 *  name : filename
 *  format : type of tape image format
 *    TAPE_RUNLENGTH  : run length encoded format (rlf)
 *    TAPE_FILE2000B : mz2000/2200 (2000BPS) format
 *    TAPE_DATA2000B : data block (binary) only format
 *    TAPE_INFO2000B : info block only format
 */

TAPE *twopen(char *name, int format)
{
	FILE *fp;
	TAPE *tp;
	char header[MAXHEADERSIZE];
	int header_length;

	if (!validwriteformat(format)) {
		return NULL;
		fprintf(stderr, "Invalid write format (%d)\n", format);
	}

	if ((fp = fopen(name, "w")) == NULL) return NULL;
	header_length = tmake_header(header, format);
	if (header_length > 0) fwrite(header, 1, header_length, fp);

	tp = tpinit(format, 1, fp, NULL);
	return tp;
}

/*
 * TAPE *twopen_mem(unsigned char *buffer, int bufsize, int format)
 *  Opening a new tape image in memory for writing.
 *  buffer : start addr of memory
 *  bufsize : max size of memory
 *  format : type of tape image format
 *    TAPE_RUNLENGTH : run length encoded format (rlf)
 *    TAPE_FILE2000B : mz2000/2200 (2000BPS) format
 *    TAPE_DATA2000B : data block (binary) only format
 *    TAPE_INFO2000B : info block only format
 */
TAPE *twopen_mem(unsigned char *buffer, int bufsize, int format)
{
	TBUFFER *bp;
	TAPE *tp;

	if (!validwriteformat(format)) {
		fprintf(stderr, "Invalid write format (%d)\n", format);
		return NULL;
	}

	tmake_header(buffer, format);
	bp = tbopen(buffer, bufsize);
	if (bp == NULL) return NULL;
	tp = tpinit(format, 1, NULL, bp);
	if (tp == NULL) {
		tbclose(bp);
		return NULL;
	}
	
	return tp;
}

/*
 *  static TAPE *tpinit(TAPE *tp, int format, iswrite, isfile)
 *
 *  Initializing tape handle.
 */
static int dummy_func(TAPE *tp)
{
	return 1;
}

#define tpinit_2000b(tp) {\
	(tp)->shortclk = 666;\
	(tp)->longclk = 1334;\
	(tp)->checkclk = 1000;\
	}

static TAPE *tpinit(int format, int iswrite, FILE *fp, TBUFFER *bp)
{
	TAPE *tp;
	int isfile = (fp != NULL);

	if ((iswrite)&&(!validwriteformat(format))) return NULL;
	if ((!iswrite)&&(!validreadformat(format))) return NULL;
	tp = (TAPE *)malloc(sizeof(TAPE));
	tp->fp = (isfile)? fp : NULL;
	tp->bp = (isfile)? NULL : bp;
	tsetflags(tp, format, iswrite, isfile);

	switch(format) {
	case TAPE_RUNLENGTH:
		tp->mainblk_offset = 0x08;
		tp->datablk_offset = 0x08;
		break;
	case TAPE_FILE2000B:
		tp->mainblk_offset = 0x08;
		tp->datablk_offset = 0x8a;
		tpinit_2000b(tp);
		tp->with_checksum = 1;
		break;
	case TAPE_DATA2000B:
		tp->mainblk_offset = -1;
		tp->datablk_offset = 0;
		tpinit_2000b(tp);
		tp->with_checksum = 0;
		break;
	case TAPE_INFO2000B:
		tp->mainblk_offset = 0;
		tp->datablk_offset = -1;
		tpinit_2000b(tp);
		tp->with_checksum = 0;
		break;
	case TAPE_BOTH2000B:
		tp->mainblk_offset = 0;
		tp->datablk_offset = 0x80;
		tpinit_2000b(tp);
		tp->with_checksum = 0;
		break;
	case TAPE_NONE2000B:
		tp->mainblk_offset = -1;
		tp->datablk_offset = -1;
		tpinit_2000b(tp);
		tp->with_checksum = 0;
		break;
	}

	if ((format == TAPE_FILE2000B) && (!iswrite)) {
		tseek(tp, 0x12, TSEEK_INFOBLK);
		tp->datasize = getword_le(tp);
		if (tp->datasize == -1) {
			free(tp);
			return NULL;
		}
#ifdef TRACE
		fprintf(stderr, "R:Size of datablock = %04x\n", tp->datasize);
#endif
	}

	trewind(tp);

	return tp;
}

/*
 *  static int tscan_header(unsigned char *header)
 *  read header information from memory
 *  return value : format code
 */
static int tscan_header(unsigned char *header)
{
	if (memcmp(header, "mz20", 4)) return 0;
	if (header[4] != 0x00) return 0;
	if (!validreadformat(header[5])) return 0;

	return header[5];
}

/*
 *  static int tmake_header(unsigned char *header, int format)
 *  write header information to memory
 *  format : format code
 *  return value : header length
 */
static int tmake_header(unsigned char *header, int format)
{
	int length = 0;
	switch(format) {
	case TAPE_RUNLENGTH:
	case TAPE_FILE2000B:
		memset(header, 0, 8);
		memcpy(header, "mz20", 4);
		header[5] = format;
		length = 8;
		break;
	}
	return length;
}

/*
 *  static int tseek(TAPE *tp, int offset, int whence)
 *  seek for tape hundle
 */
static int tseek(TAPE *tp, int offset, int whence)
{
	int base = -1;

	switch(whence) {
	case TSEEK_HEADER:
		break;
	case TSEEK_MAINBLK:
	case TSEEK_INFOBLK:
		base = tp->mainblk_offset;
		break;
	case TSEEK_DATABLK:
		base = tp->datablk_offset;
		break;
	default:
		break;
	}
	if (base < 0) {
		tsetnowrite(tp);
	} else {
		tresetnowrite(tp);
		tflush(tp);
		if (tisfile(tp)) {
			fseek(tp->fp, base + offset, SEEK_SET);
		} else {
			tp->bp->curoffset = base + offset;
		}
	}
	return 0;
}

/*
 * int tflush(TAPE *tp)
 * flushing the tape handle
 */
static int tflush(TAPE *tp)
{
	if (tisfile(tp)) fflush(tp->fp);
	return 0;
}

/*
 * int tclose(TAPE *tp);
 *  Closing the tape handle;
 *  tp : tape handle;
 */
int tclose(TAPE *tp)
{
	int size;
	if (tiswrite(tp)) twrite(tp, 0, 1);
	if (!tisfinish(tp)) fprintf(stderr, "tape closed without finishing\n");
	if (tisfile(tp)) {
		size = ftell(tp->fp);
		fclose(tp->fp);
	} else {
		size = tp->bp->curoffset;
		tbclose(tp->bp);
	}
	if (tiserror(tp)) size = -1;
	free(tp);
	return size;
}

/*
 * int trewind(TAPE *tp)
 *  rewind to top of the tape.
 *  tp : tape handle
 */
int trewind(TAPE *tp)
{
	tp->state = tp->hl = tp->clock = 0;
	tp->skip_func = dummy_func;
	tp->write_func = dummy_func;
	tresetfinish(tp);
	tseek(tp, 0, TSEEK_MAINBLK);
	return 0;
}

/*
 * int tscan(TAPE *tp);
 *  Scanning the current position in the tape, and telling high or low.
 *  tp : tape handle;
 */
int tscan(TAPE *tp)
{
	return(tp->hl);
}

/*
 *  functions for tskip
 */

/*
 *  skip_????(TAPE *tp);
 *   subfunctions for tskip(tp);
 *  skip_????_init(TAPE *tp, ...);
 *   initialize for skip_???? function
 *
 *  skip_blank: skipping in blank phase(pre-tape/post-tape)
 *  skip_tapeend: skipping after reaching end of tape
 *  skip_rlbody: skipping for runlength format
 *  skip_repeat: skipping in repeating to toggle high and low
 *  skip_mzdata: skipping for inf/data block of mz2000/2200 format
 */

#define  SUBMIN(a, b) {if (a <= b) {b -= a; a = 0;} else {a -= b; b = 0;}}

static int skip_blank(TAPE *tp)
{
	SUBMIN(tp->u.skblank.restclock, tp->clock);
	return (tp->u.skblank.restclock == 0)? 1 : 0;
}

static int skip_blank_init(TAPE *tp, int period)
{
	tp->skip_func = skip_blank;
	tp->u.skblank.restclock = period;
	tp->hl = 0;
	return 0;
}

static int skip_rlbody(TAPE *tp)
{
	int retval = 0;

	while (tp->clock > 0) {
		if (tp->u.skrlbody.restclock == 0) {
			z80clk x = getclk(tp);
			if (x == -1) {
				retval = 1;
				break;
			}
			tp->u.skrlbody.restclock = x;
			tp->hl = (tp->hl == 0)? 1 : 0;
		}
		SUBMIN(tp->u.skrlbody.restclock, tp->clock);
	}
	return retval;
}

static int skip_rlbody_init(TAPE *tp)
{
	tp->skip_func = skip_rlbody;
	tp->u.skrlbody.restclock = 0;
	tp->hl = 0;
	return 0;
}

static int skip_tapeend(TAPE *tp)
{
	tp->clock = 0;
	return 0;
}

static int skip_tapeend_init(TAPE *tp)
{
	tp->skip_func = skip_tapeend;
	tp->hl = -1;
	return 0;
}

static int skip_repeat(TAPE *tp)
{
	int retval = 0;
	struct st_skrepeat *local = &(tp->u.skrepeat);

	while (tp->clock > 0) {
		if (local->restclock == 0) {
			if (tp->hl == 0) {
				tp->hl = 1;
			} else {
				tp->hl = 0;
				if (--local->restrep == 0) {
					retval = 1;
					break;
				}
			}
			local->restclock += (local->replong)?
				tp->longclk : tp->shortclk;
		}
		SUBMIN(local->restclock, tp->clock);
	}
	return retval;
}

static int skip_repeat_init(TAPE *tp, int nrepeats, int replong)
{
	struct st_skrepeat *local = &(tp->u.skrepeat);

	tp->skip_func = skip_repeat;
	local->restrep = nrepeats;
	local->replong = replong;
	local->restclock = (replong)? tp->longclk : tp->shortclk;
	tp->hl = 0;
	return 0;
}

static int skip_mzdata(TAPE *tp)
{
	int retval = 0;
	int data = 0;
	struct st_skmzdata *local = &(tp->u.skmzdata);

	while (tp->clock > 0) {
		if (local->restclock == 0) {
			if (tp->hl == 0) {
				tp->hl = 1;
			} else {
				tp->hl = 0;
				if (local->curbit-- <= 0) {
					local->restndata--;
					if (local->restndata == -3) {
						int checksum;
						if (!tp->with_checksum) {
							retval = 1;
							break;
						}
						checksum = getword_be(tp);
						if (checksum == (local->checksum&0xffff)) {
							retval = 1;
						} else {
							fprintf(stderr,"R:%04x:%04x\n",
								local->checksum & 0xffff,
								checksum);
							retval = -1;
						}
						break;
					} else if (local->restndata == -1) {
						data = (local->checksum >> 8) & 0xff;
					} else if (local->restndata == -2) {
						data = (local->checksum & 0xff);
					} else {
						data = getbyte(tp);
						local->checksum += count_nbits(data);
					}
					if (data == -1) {
						retval = -1;
						break;
					}
					local->curdata = data;
					local->curbit = 8;
					local->islong = 1;
				} else {
					local->islong = ((local->curdata) >> local->curbit) & 0x01;
				}
			}
			local->restclock += (local->islong)? tp->longclk : tp->shortclk;
		}
		SUBMIN(local->restclock, tp->clock);
	}
	return retval;
}

static int skip_mzdata_init(TAPE *tp, int byte)
{
	struct st_skmzdata *local = &(tp->u.skmzdata);

	tp->skip_func = skip_mzdata;
	local->restclock = 0;
	local->restndata = byte;
	local->curbit = 0;
	local->checksum = 0;
	tp->hl = 1;
	return 0;
}

static void tskip_next_state_runlength(TAPE *tp)
{
	switch (tp->state) {
	case 0:
		skip_blank_init(tp, 16000000);  /* 4sec */
		break;
	case 1:
		tseek(tp, 0, TSEEK_MAINBLK);
		skip_rlbody_init(tp);
		break;
	case 2:
		tsetfinish(tp);
		skip_blank_init(tp, 16000000);  /* 4sec */
		break;
	default:
		skip_tapeend_init(tp);
		break;
	}
	tp->state++;
}

static void tskip_next_state_file2000b(TAPE *tp)
{
	switch (tp->state) {
	case 0:
#ifdef TRACE
		fprintf(stderr, "R:pre-gap(4sec)\n");
#endif
		skip_blank_init(tp, 16000000);  /* 4sec */
		break;
	case 1:
#ifdef TRACE
		fprintf(stderr, "R:long mark for info block\n");
#endif
		skip_repeat_init(tp, 0x2710, 0);
		break;
	case 2:
#ifdef TRACE
		fprintf(stderr, "R:short mark1 for info block\n");
#endif
		skip_repeat_init(tp, 40, 1);
		break;
	case 3:
#ifdef TRACE
		fprintf(stderr, "R:short mark2 for info block\n");
#endif
		skip_repeat_init(tp, 40, 0);
		break;
	case 4:
		skip_repeat_init(tp, 1, 1);
		break;
	case 5:
		tseek(tp, 0, TSEEK_INFOBLK);
#ifdef TRACE
		fprintf(stderr, "R:data of info block (1st)\n");
#endif
		skip_mzdata_init(tp, 0x80);
		break;
	case 6:
		skip_repeat_init(tp, 1, 1);
		break;
	case 7:
		tseek(tp, 0, TSEEK_INFOBLK);
#ifdef TRACE
		fprintf(stderr, "R:data of info block (2nd)\n");
#endif
		skip_mzdata_init(tp, 0x80);
		break;
	case 8:
		skip_repeat_init(tp, 1, 1);
		break;
	case 9:
#ifdef TRACE
		fprintf(stderr, "R:long mark for data block\n");
#endif
		skip_repeat_init(tp, 0x2af8, 0);
		break;
	case 10:
#ifdef TRACE
		fprintf(stderr, "R:short mark1 for data block\n");
#endif
		skip_repeat_init(tp, 20, 1);
		break;
	case 11:
#ifdef TRACE
		fprintf(stderr, "R:short mark2 for data block\n");
#endif
		skip_repeat_init(tp, 20, 0);
		break;
	case 12:
		skip_repeat_init(tp, 1, 1);
		break;
	case 13:
		tseek(tp, 0, TSEEK_DATABLK);
#ifdef TRACE
		fprintf(stderr, "R:data of data block (1st)\n");
#endif
		skip_mzdata_init(tp, tp->datasize);
		break;
	case 14:
		skip_repeat_init(tp, 1, 1);
		break;
	case 15:
		tseek(tp, 0, TSEEK_DATABLK);
#ifdef TRACE
		fprintf(stderr, "R:data of data block (2nd)\n");
#endif
		skip_mzdata_init(tp, tp->datasize);
		break;
	case 16:
		skip_repeat_init(tp, 1, 1);
		break;
	case 17:
		tsetfinish(tp);
#ifdef TRACE
		fprintf(stderr, "R:post-gap(4sec)\n");
#endif
		skip_blank_init(tp, 16000000);  /* 4sec */
		break;
	case 18:
#ifdef TRACE
		fprintf(stderr, "R:read finished\n");
#endif
		break;
	default:
		skip_tapeend_init(tp);
		break;
	}
	tp->state++;
}

static int tskip_next_state(TAPE *tp)
{
	if (tformat(tp) == TAPE_RUNLENGTH) {
		tskip_next_state_runlength(tp);
	} else if (tformat(tp) == TAPE_FILE2000B) {
		tskip_next_state_file2000b(tp);
	}
	return tp->state;
}

/*
 * int tskip(TAPE *tp, z80clk clock);
 *  Skipping clocks in the tape
 *  tp : tape handle;
 *  clock : skipping clocks
 */
int tskip(TAPE *tp, z80clk clock)
{
	if ((tiserror(tp)) || (tiswrite(tp))) return -1;
	tp->clock = clock;
	while (tp->clock > 0) {
		int retval = tp->skip_func(tp);
		if (retval == -1) {
			fprintf(stderr, "R:read error\n");
			tseterror(tp);
			tp->hl = -1;
			return -1;
		}
		if (retval) tskip_next_state(tp);
	}
	return tp->hl;
}

/*
 *  functions for twrite
 */

/*
 *  write_????(TAPE *tp);
 *   subfunctions for twrite(tp);
 *  write_????_init(TAPE *tp, ...);
 *   initialize for write_???? function
 *
 *  write_blank: writing for blank phase(pre-tape/post-tape)
 *  write_rlbody: writing for runlength format
 *  write_repeat: writing in repeating to toggle high and low
 *  write_mzdata: writing for inf/data block of mz2000/2200 format
 */
static int write_blank(TAPE *tp)
{
	int retval = 0;

	if (tp->hl == 0) {
		tp->clock = 0;
	} else {
		retval = 1;
	}
	return retval;
}

static int write_blank_init(TAPE *tp)
{
	tp->write_func = write_blank;
	return 0;
}

static int write_rlbody(TAPE *tp)
{
	struct st_wrrlbody *local = &(tp->u.wrrlbody);

	if (local->prevhl != tp->hl) {
		if (putclk(local->interval, tp) == -1) {
			tseterror(tp);
			return 0;
		}
		local->prevhl = tp->hl;
		local->interval = 0;
	}
	local->interval += tp->clock;
	tp->clock = 0;
	return 0;
}

static int write_rlbody_init(TAPE *tp)
{
	struct st_wrrlbody *local = &(tp->u.wrrlbody);

	tp->write_func = write_rlbody;
	local->prevhl = 1;
	local->interval = 0;
	return 0;
}

static int write_repeat(TAPE *tp)
{
	int retval = 0;
	struct st_wrrepeat *local = &(tp->u.wrrepeat);

	switch(local->searchhi) {
	case 0:
		if (tp->hl == 0) {
			local->searchhi = 1;
		} else {
			tp->clock = 0;
		}
		break;
	case 1:
		if (tp->hl == 1) {
			local->searchhi = 2;
			local->restclock = tp->checkclk;
		} else {
			tp->clock = 0;
		}
		break;
	case 2:
		SUBMIN(local->restclock, tp->clock);
		if (local->restclock == 0) {
			if ((local->replong == tp->hl)
			    ||(local->replong == -1)) {
				if (--local->restrep == 0) {
					retval = 1;
				}
			} else {
				retval = -1;
			}
			local->searchhi = 0;
		}
		break;
	}
	return retval;
}

static int write_repeat_init(TAPE *tp, int nrepeats, int replong)
{
	struct st_wrrepeat *local = &(tp->u.wrrepeat);

	tp->write_func = write_repeat;
	local->searchhi = 0;
	local->restrep = nrepeats;
	local->replong = replong;
	return 0;
}

static int write_mzdata(TAPE *tp)
{
	int retval = 0;
	struct st_wrmzdata *local = &(tp->u.wrmzdata);

	switch(local->searchhi) {
	case 0:
		if (tp->hl == 0) {
			local->searchhi = 1;
		} else {
			tp->clock = 0;
		}
		break;
	case 1:
		if (tp->hl == 1) {
			local->searchhi = 2;
			local->restclock = tp->checkclk;
		} else {
			tp->clock = 0;
		}
		break;
	case 2:
		SUBMIN(local->restclock, tp->clock);
		if (local->restclock == 0) {
			if (++local->curbit == 9) {
				local->curdata &= 0xff;
				if ((local->send < local->size)
				    ||(tp->with_checksum)) {
					if (putbyte(local->curdata, tp) == -1) {
						tseterror(tp);
						break;
					}
				}
				local->send++;
				if (local->send == 0x13)
					local->datasize = local->curdata;
				if (local->send == 0x14) {
					local->datasize += local->curdata * 256;
				}
				if (local->send == local->size + 1)
					local->checksum = local->curdata * 256;
				if (local->send == local->size + 2) {
					local->checksum += local->curdata;
					if ((local->nbits & 0xffff) == local->checksum) {
						retval = 1;
					} else {
						fprintf(stderr,"%04x : %04x\n",
							local->nbits & 0xffff,
							local->checksum);
						retval = -1;
					}
					break;
				}
				local->curbit = 0;
				local->curdata = 0;
			} else {
				local->curdata <<= 1;
				local->curdata |= tp->hl;
				if ((local->send < local->size)&&
				    (tp->hl)) local->nbits += 1;
			}
			local->searchhi = 0;
		}
		break;
	}
	return retval;
}

static int write_mzdata_init(TAPE *tp, int byte)
{
	struct st_wrmzdata *local = &(tp->u.wrmzdata);

	tp->write_func = write_mzdata;
	local->searchhi = 0;
	local->restclock = 0;
	local->size = byte;
	local->send = 0;
	local->curbit = 0;
	local->curdata = 0;
	local->nbits = 0;
	return 0;
}

static int write_tapeend(TAPE *tp)
{
	tp->clock = 0;
	return 0;
}

static int write_tapeend_init(TAPE *tp)
{
	tp->write_func = write_tapeend;
	return 0;
}

static void twrite_next_state_runlength(TAPE *tp)
{
	switch(tp->state) {
	case 0:  /* Ignoring first low-level interval */
		write_blank_init(tp);
		break;
	case 1:  /* Writing the number of clocks of high/low-level interval */
		tseek(tp, 0, TSEEK_MAINBLK);
		tsetfinish(tp);
		write_rlbody_init(tp);
		break;
	}
	tp->state++;
}

static void twrite_next_state_file2000b(TAPE *tp)
{
	switch(tp->state) {
	case 0: /* Detecting first mark for info block (fail->state0)*/
		write_repeat_init(tp, 40, 1);
		break;
	case 1: /* Detecting second mark for info block (fail->state0)*/
		write_repeat_init(tp, 40, 0);
		break;
	case 2: /* Detecting one wave */
#ifdef TRACE
		fprintf(stderr, "W:Found mark of info block\n");
#endif
		tp->nfail = 0;
		write_repeat_init(tp, 1, -1);
		break;
	case 3: /* Detecting one long wave (fail->error) */
		write_repeat_init(tp, 1, 1);
		break;
	case 4: /* Writing an info block (fail->(1st)state3(2nd)error) */
		tseek(tp, 0, TSEEK_INFOBLK);
#ifdef TRACE
		fprintf(stderr, "W:Now writing info block\n");
#endif
		write_mzdata_init(tp, 0x80);
		break;
	case 5: /* Dummy state */
		tp->write_func = dummy_func;
		tp->datasize = tp->u.wrmzdata.datasize;
#ifdef TRACE
		fprintf(stderr, "W:Size of datablock = %04x\n", tp->datasize);
#endif
		break;
	case 6: /* Detecting first mark for data block (fail->state6) */
		write_repeat_init(tp, 20, 1);
		break;
	case 7: /* Detecting second mark for data block (fail->state6) */
		write_repeat_init(tp, 20, 0);
		break;
	case 8: /* Detecting one wave */
#ifdef TRACE
		fprintf(stderr, "W:Found mark of data block\n");
#endif
		tp->nfail = 0;
		write_repeat_init(tp, 1, -1);
		break;
	case 9: /* Detecting one long wave (fail->error) */
		write_repeat_init(tp, 1, 1);
		break;
	case 10: /* Writing a data block (fail->(1st)state9(2nd)error) */
		tseek(tp, 0, TSEEK_DATABLK);
#ifdef TRACE
		fprintf(stderr, "W:Now writing data block\n");
#endif
		write_mzdata_init(tp, tp->datasize);
		break;
	case 11: /* Success */
#ifdef TRACE
		fprintf(stderr, "W:Write Success!\n");
#endif
		tflush(tp);
		tsetfinish(tp);
	default:
		write_tapeend_init(tp);
		break;
	}
	tp->state++;
}

static void twrite_fail_state_file2000b(TAPE *tp)
{
	switch(tp->state) {
	case 1: case 2:
		tp->state = 0;
		break;
	case 7: case 8:
		tp->state = 6;
		break;
	case 4: case 10:
		tp->state--;
		break;
	case 5: case 11:
		if (++tp->nfail == 1) {
			fprintf(stderr, "W:Checksum mismatch (1st) in %s\n",
				(tp->state==5)? "info block" : "data block");
			tp->state -= 2;
		} else {
			fprintf(stderr, "W:Checksum mismatch (2nd) in %s\n",
				(tp->state==5)? "info block" : "data block");
			tseterror(tp);
		}
		break;
	default:
		fprintf(stderr, "Write failure in %d state\n", tp->state);
		tseterror(tp);
		break;
	}
}

static int twrite_next_state(TAPE *tp)
{
	if (tformat(tp) == TAPE_RUNLENGTH) {
		twrite_next_state_runlength(tp);
	} else if ((tformat(tp) == TAPE_FILE2000B)
		|| (tformat(tp) == TAPE_DATA2000B)
		|| (tformat(tp) == TAPE_INFO2000B)
		|| (tformat(tp) == TAPE_BOTH2000B)
		|| (tformat(tp) == TAPE_NONE2000B)) {
		twrite_next_state_file2000b(tp);
	}
	return tp->state;
}

static int twrite_fail_state(TAPE *tp)
{
	if ((tformat(tp) == TAPE_FILE2000B)
	    || (tformat(tp) == TAPE_DATA2000B)
	    || (tformat(tp) == TAPE_INFO2000B)
	    || (tformat(tp) == TAPE_BOTH2000B)
	    || (tformat(tp) == TAPE_NONE2000B)) {
		twrite_fail_state_file2000b(tp);
	}
	return tp->state;
}

/*
 * int twrite(TAPE *tp, int hl, z80clk clock);
 *  Writing high/low data in some clocks
 *  tp : tape handle;
 *  hl : written data is high or not
 *  clock : writing clocks
 */
int twrite(TAPE *tp, int hl, z80clk clock)
{
	if ((tiserror(tp)) || (!tiswrite(tp))) return -1;
	tp->clock = clock;
	tp->hl = (hl == 0)? 0 : 1;
	while (tp->clock > 0) {
		int retval = tp->write_func(tp);
		if (retval == -1) twrite_fail_state(tp);
		if (tiserror(tp)) return -1;
		if (retval) twrite_next_state(tp);
	}
	return 1;
}

int tcopy(TAPE *intp, TAPE *outtp, int step)
{
	int c, prev = 0, clock = 0;
	trewind(intp);
	trewind(outtp);
	while (1) {
		c = tskip(intp, step);
		if (c == prev) {
			clock += step;
		} else {
			twrite(outtp, prev, clock);
			prev = c;
			clock = step;
		}
		if (c == -1) break;
	}
	return 1;
}

/*
 * int tcheck_tape(TAPE *intp)
 *  checking whether tape is valid
 *  intp: checked tape handle
 */

int tcheck_tape(TAPE *intp)
{
	TAPE *outtp;
	int retval = 0;

	trewind(intp);
	if ((outtp = twopen_mem(NULL, 0, TAPE_NONE2000B)) == NULL) {
		fprintf(stderr, "Unexpected error occured\n");
		return 0;
	}
	tcopy(intp, outtp, 90);
	if ((tisfinish(intp))&&(!tiserror(intp))
	    &&(tisfinish(outtp))&&(!tiserror(outtp))) retval = 1;
	tclose(outtp);
	return retval;
}
