#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mz2000.h"

#define fail_return(v) {pthread_mutex_unlock(&(mz2000->tdeck_mutex));return v;}


int init_tdeck(MZ2000 *mz2000)
{
	MZTDeck *tdeck;

	if ((tdeck = (MZTDeck *)calloc(sizeof(MZTDeck),1)) == NULL) {
		fprintf(stderr, "No enough memory\n");
		return 0;
	}
	mz2000->tdeck = tdeck;
	pthread_mutex_init(&(mz2000->tdeck_mutex), NULL);
	tdeck->status = TDECK_EJECTED;
	tdeck_clear(mz2000);
	return 1;
}

int tdeck_status(MZ2000 *mz2000)
{
	int retval;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	retval = mz2000->tdeck->status;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return retval;
}

int tdeck_slotused(MZ2000 *mz2000)
{
	int retval;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	retval = mz2000->tdeck->used;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return retval;
}

int tdeck_stop(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->tp != NULL) {
		tclose(tdeck->tp);
		tdeck->tp = NULL;
		tdeck->current++;
	}
	tdeck->status = TDECK_STOPPED;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_eject(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status == TDECK_EJECTED) fail_return(0);
	if (tdeck->status != TDECK_STOPPED) tdeck_stop(mz2000);
	tdeck->status = TDECK_EJECTED;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_set(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_EJECTED) fail_return(0);
	tdeck->status = TDECK_STOPPED;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_ff(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status == TDECK_EJECTED) fail_return(0);
	if (tdeck->status != TDECK_STOPPED) tdeck_stop(mz2000);
	if (tdeck->current < tdeck->used) tdeck->current++;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_rew(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status == TDECK_EJECTED) fail_return(0);
	if (tdeck->status != TDECK_STOPPED) tdeck_stop(mz2000);
	if (tdeck->current > 0) tdeck->current--;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_clear(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	int i;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_EJECTED) fail_return(0);

	for (i = 0; i < TDECK_MAX_SLOTS; i++) {
		tdeck->slots[i].filename = NULL;
		tdeck->slots[i].format = 0;
	}

	tdeck->used = tdeck->current = 0;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_insert(MZ2000 *mz2000, char *filename, int format, int slot)
{
	MZTDeck *tdeck;
	int i;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_EJECTED) fail_return(0);
	if (tdeck->used >= TDECK_MAX_SLOTS) fail_return(0);
	if ((slot < 0)||(slot > tdeck->used)) fail_return(0);
	for (i = tdeck->used; i >= slot; i--) 
		memcpy(&(tdeck->slots[i+1]), &(tdeck->slots[i]), sizeof(struct mztdeck_slotinfo));
	tdeck->slots[slot].filename = strdup(filename);
	tdeck->slots[slot].format = format;
	tdeck->used++;
	if (tdeck->current > slot) tdeck->current++;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_delete(MZ2000 *mz2000, int slot)
{
	MZTDeck *tdeck;
	int i;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_EJECTED) fail_return(0);
	if (tdeck->used <= 0) fail_return(0);
	if ((slot < 0)||(slot > tdeck->used)) fail_return(0);
	free(tdeck->slots[slot].filename);
	for (i = slot; i < tdeck->used - 1; i++) 
		memcpy(&(tdeck->slots[i]), &(tdeck->slots[i+1]), sizeof(struct mztdeck_slotinfo));
	tdeck->slots[tdeck->used - 1].filename = NULL;
	tdeck->slots[tdeck->used - 1].format = 0;
	tdeck->used--;
	if (tdeck->current > slot) tdeck->current--;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_add(MZ2000 *mz2000, char *filename, int format)
{
	return tdeck_insert(mz2000, filename, format, mz2000->tdeck->used);
}

static int tdeck_open(MZ2000 *mz2000, int iswrite)
{
	MZTDeck *tdeck;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_STOPPED) fail_return(0);
	while (tdeck->current < tdeck->used) {
		if (iswrite) {
			tdeck->tp = twopen(tdeck->slots[tdeck->current].filename, tdeck->slots[tdeck->current].format);
		} else {
			tdeck->tp = tropen(tdeck->slots[tdeck->current].filename);
		}
		if (tdeck->tp != NULL) break;
		fprintf(stderr, "%s is invalid filename\n",
			tdeck->slots[tdeck->current].filename);
		tdeck_delete(mz2000, tdeck->current);
	}
	if (tdeck->tp == NULL) fail_return(0);
	tdeck->status = (iswrite)? TDECK_WRITING : TDECK_READING;
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return 1;
}

int tdeck_ropen(MZ2000 *mz2000)
{
	return tdeck_open(mz2000, 0);
}

int tdeck_wopen(MZ2000 *mz2000)
{
	return tdeck_open(mz2000, 1);
}

int tdeck_skip(MZ2000 *mz2000, z80clk clock)
{
	MZTDeck *tdeck;
	int retval;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_READING) fail_return(-1);
	retval = tskip(tdeck->tp, clock);
	if (retval == -1) tdeck_stop(mz2000);
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return retval;
}

int tdeck_scan(MZ2000 *mz2000)
{
	MZTDeck *tdeck;
	int retval;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_READING) fail_return(-1);
	retval = tscan(tdeck->tp);
	if (retval == -1) tdeck_stop(mz2000);
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return retval;
}

int tdeck_write(MZ2000 *mz2000, int bit, z80clk clock)
{
	MZTDeck *tdeck;
	int retval;

	pthread_mutex_lock(&(mz2000->tdeck_mutex));
	tdeck = mz2000->tdeck;
	if (tdeck->status != TDECK_WRITING) fail_return(-1);
	retval = twrite(tdeck->tp, bit, clock);
	if (retval == -1) tdeck_stop(mz2000);
	pthread_mutex_unlock(&(mz2000->tdeck_mutex));
	return retval;
}
